#include "StdAfx.h"
#include "HeightField.h"


HeightField::HeightField(int x, int y, int z)
{
	this->columnsX = x + 2;
	this->columnsY = y + 2;

	heightField = (float**) malloc( sizeof(float*) * columnsX );
	for (int i = 0; i < columnsX; i++) {
		heightField[i] = (float*) malloc( sizeof(float) * columnsY );
	}

	velocityField = (float**) malloc( sizeof(float*) * columnsX );
		for (int i = 0; i < columnsX; i++) {
		velocityField[i] = (float*) malloc( sizeof(float) * columnsY );
	}

	forceField = (float**) malloc( sizeof(float*) * columnsX );
		for (int i = 0; i < columnsX; i++) {
		forceField[i] = (float*) malloc( sizeof(float) * columnsY );
	}


	this->initialHeight = z;
	for (int i = 0; i < columnsX; i++) {
		for (int j = 0; j < columnsY; j++) {
			heightField[i][j] = initialHeight;
			velocityField[i][j] = 0;
			forceField[i][j] = 0;
		}
	}
	
	//velocityField[ ((int)columnsX / 2)  ][ ((int)columnsY / 2) ] = 2 * (initialHeight / 3);


	//heightField[0][0] = 1.0;
	//heightField[0][1] = 1.0;

	//velocityField[0][0] = -10.0;
	//velocityField[0][1] = -10.0;

	//forceField[0][0] = 10.0;
	//forceField[0][1] = 10.0;

	// hf[0,0] 1 
	// hf[0,1] 0.5
	// vf[0,0] -1
	// vf[0,1] -0.5
	// ff[0,0] 10
	// ff[0,1] -10

	time = 0;
	
	waveSpeedSquared = 8; // Should be < 10
	//
	// NOTE: If the timestep is high (numerically explosive) then the velocity drag must 
	//       be equally high (closer to one) so stability can be reached!
	//       vd can be 33 times timestep
	timeStep = 0.03; // Works better if less than .05
	velocityDrag = .992; // Drag must be proportional to timeStep 
	//time 0, wavespeed 8 timestep .03 velocitydrag .99

	std::cout << "Initial Height: " << heightField[0][0] << std::endl;
}


HeightField::~HeightField(void)
{
	for (int i = 0; i < columnsX; i++) {
		delete heightField[i];
	}
}

bool HeightField::loadTextures() {
	GLuint numberOfColors;
	GLenum textureFormat;

	textureSurface = SDL_LoadBMP("WaterTexture.bmp");

	if (!textureSurface) {
			std::cout << "Couldn't Load Texture." << std::endl;
		return false;;
	}

	numberOfColors = textureSurface->format->BytesPerPixel;
	switch(numberOfColors) {
	case(4):
			textureFormat = GL_BGRA;
		break;
	case(3):
			textureFormat = GL_BGR;
		break;
	}

	glGenTextures(1, &textureHandle);
	glBindTexture( GL_TEXTURE_2D, textureHandle );

	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP); // GL_REFLECTION_MAP

	glTexImage2D( GL_TEXTURE_2D, 0, numberOfColors, textureSurface->w, textureSurface->h, 0, textureFormat, GL_UNSIGNED_BYTE, textureSurface->pixels );

	return true;
}

void HeightField::setTimeStep(float lessThanOne) {
	this->timeStep = lessThanOne;
	this->velocityDrag = timeStep * 33.07;
	this->waveSpeedSquared = timeStep * 266.66; //333.3;
}

void HeightField::splashAtPoint(int x, int y, float height) {
	if (x > 2 && x < columnsX - 1 && y > 2 && y < columnsY - 1) {
		for (int i = -1; i < 1; i++) {
			for (int j = -1; j < 1; j++) {
				if ( i != -1 || j != -1 || i != 1 || j != 1 ) {
					
					//heightField[x + i][y + j] = (initialHeight - height);
					//heightField[x + i][y + j] -= 1;
					velocityField[x + i][y + j] -= height;

				//} else {
					//heightField[x + i][y + j] = initialHeight + (initialHeight - height);
				}
			}
		}
		//heightField[x][y] = height;
		//heightField[ x  + 1 ][ y + 1 ] = 2 * height/3;
		//heightField[ x - 1 ][ y + 1 ] = 2 * height/3;
		//heightField[ x - 1 ][ y - 1 ] = 2 * height/3;
		//heightField[ x + 1 ][ y - 1 ] = 2 * height/3;
	}
}

void HeightField::splashCenter(float height) {
	splashAtPoint( (columnsX - 2) / 2, (columnsY - 2) / 2, height);
	//heightField[ (columnsX / 2) ][ (columnsY / 2) ] = height;
	//heightField[ ((int)columnsX / 2)  + 1 ][ ((int)columnsY / 2) + 1 ] = 2 * height / 3;
	//heightField[ ((int)columnsX / 2) - 1 ][ ((int)columnsY / 2) + 1 ] = 2 * height / 3;
	//heightField[ ((int)columnsX / 2) - 1 ][ ((int)columnsY / 2) - 1 ] = 2 * height / 3;
	//heightField[ ((int)columnsX / 2) + 1 ][ ((int)columnsY / 2) - 1 ] = 2 * height / 3;
}

void HeightField::update() {

	float sum = 0.0;
	float avg = 0.0;

	for (int x = 1; x < columnsX - 1; x++) { // LESS THAN
		for (int y = 1; y < columnsY - 1; y++) {

			//
			// If time % 1 = non deterministic ...
			if (time % 2 == 0) {
				forceField[x][y] = heightField[x + 1][y] + heightField[x - 1][y] + heightField[x][y + 1] + heightField[x][y - 1] - 4 * heightField[x][y];
				forceField[x][y] = waveSpeedSquared * forceField[x][y] / 1.0;

				velocityField[x][y] = velocityField[x][y] + forceField[x][y] * timeStep; // Time step of 0.01
				//velocityField[x][y] = velocityField[x][y] * 0.999999;

				heightField[x][y]  = heightField[x][y] + velocityField[x][y] * timeStep;
				//heightField[x][y] = heightField[x][y] + ( ( heightField[x][y] - initialHeight ) / 5 );
				// 
				// Water can't go through the ground!
				//if (heightField[x][y] < 0.1 * initialHeight) {
					//heightField[x][y] = 0.1 * initialHeight;
				//}

				sum += heightField[x][y];
			}

			avg = sum / ( (columnsX - 3) * (columnsY - 3) );
		}
	}

		//
	// Get the edges
	//for (int i = 0; i < columnsX - 1; i++) {
		//velocityField[i][0] = velocityField[i][1];
		//velocityField[i][columnsY - 1] = velocityField[i][columnsY - 2];
		//velocityField[0][i] = velocityField[1][i];
		//velocityField[columnsX - 1][i] = velocityField[columnsX - 2][i];

		//heightField[i][0]  = heightField[i][1];
		//heightField[i][columnsY - 1] = heightField[i][columnsY - 2];
		//heightField[0][i] = heightField[1][i];
		//heightField[columnsX -1][i] = heightField[columnsX - 2][i];

		//heightField[i][0]  = heightField[i][0] + velocityField[i][0] * 0.05;
		//heightField[i][columnsY - 1] = heightField[i][columnsY - 1] + velocityField[i][columnsY - 1] * 0.05;
		//heightField[0][i] = heightField[0][i] + velocityField[0][i] * 0.05;
		//heightField[columnsX -1][i] = heightField[columnsX - 1][i] + velocityField[columnsX - 1][i] * 0.05;
	//}


	//
	// Clamping the edges
	//for (int i = 0; i < columnsX; i++) {
		//for (int j = 0; j < columnsY - 1; j++) {

			//heightField[i][j] = ( heightField[i][j] + initialHeight ) / 2;
			//velocityField[i][j] = velocityField[i][j] / 2;
			//forceField[i][j] = forceField[i][j] / 2;
		//}
		//heightField[i][0] = heightField[i][1] + (heightField[i][1] - heightField[i][2]) + ((heightField[i][1] - heightField[i][2]) - (heightField[i][2] - heightField[i][3]));
		//heightField[i][columnsY  - 1] = heightField[i][columnsY - 2];
	//}
	//for (int i = 0; i < columnsY; i++) {
		//heightField[0][i] = heightField[1][i];
		//heightField[columnsX - 1][i] = heightField[columnsX - 2][i];
	//}

	//
	// Damp all velocities
	for (int x = 1; x < columnsX - 1; x++) {
		for (int y = 1; y < columnsY - 1; y++) {
			velocityField[x][y] = velocityDrag * velocityField[x][y];
			//heightField[x][y] = (heightField[x][y] + initialHeight) / 2;
		}
	}


	// 
	// Sleep the water
	//if ( abs( avg - (float)initialHeight ) < 0.0001 ) {
		//for (int x = 0; x < columnsX - 1; x++) {
		//	for (int y = 0; y < columnsY - 1; y++) {
		//		heightField[x][y] = ( (heightField[x][y] + initialHeight) / 2 ) ;
		//		velocityField[x][y] /= 2;
		//	}
		//}
	//}

	if (time % 500 == 0) {
			std::cout << "hf[1,1]: " << heightField[1][1] << " @ " << time << std::endl;
			std::cout << "ff[1,1]: " << forceField[1][1] << std::endl;
			std::cout << "vf[1,1]: " << velocityField[1][1] << std::endl;
			std::cout << "avg height: " << avg << std::endl;
	}

	time += 1;
}

void HeightField::drawSimple() {
	glColor3f(0.2f, 0.2f, 0.2f);
	glBegin(GL_QUADS);
	
	glNormal3f(0.0f, 0.0f, 1.0f);
	glVertex3f(0,0,0);

	glNormal3f(0.0f, 0.0f, 1.0f);
	glVertex3f(0,columnsY,0);
	
	glNormal3f(0.0f, 0.0f, 1.0f);
	glVertex3f(columnsX,columnsY,0);

	glNormal3f(0.0f, 0.0f, 1.0f);
	glVertex3f(columnsX,0,0);
	glEnd();
	
	for (int x = 0; x < columnsX - 1; x++) {
		for (int y = 0; y < columnsY - 1; y++) {

			//glBegin(GL_LINES);
			//glColor3f(0.1f, 0.1f, 0.3f);
			//glVertex3f(x, y, 0);
			//glVertex3f(x, y, heightField[x][y] - 2);
			//glEnd();

			if (forceField[x][y] > 0) { // Use to be velocity colored ForceField is -180 out of phase so switch signs! 
				glColor3f(0.3f, 0.2f, 1.0f);  // Valleys are blue and deceleration (mountains) or white
			} else if (forceField[x][y] < 0) {
				glColor3f(0.9f, 0.9f, 1.0f);
			} else {
				glColor3f(0.5f, 0.5f, 0.5f);
			}
			//if (heightField[x][y] == initialHeight) {
				//glColor3f(0.5f, 0.5f, 0.5f);
			//}
			
			//glBegin(GL_LINES);
			//glVertex3f(x, y, heightField[x][y] - 2);
			//glVertex3f(x, y, heightField[x][y]);	
			//glEnd();
		}
	}


}

void HeightField::drawGround() {
	glColor3f(0.2f, 0.2f, 0.2f);
	glBegin(GL_QUADS);
	
	glNormal3f(0.0f, 0.0f, 1.0f);
	glVertex3f(0,0,0);

	glNormal3f(0.0f, 0.0f, 1.0f);
	glVertex3f(0,columnsY,0);
	
	glNormal3f(0.0f, 0.0f, 1.0f);
	glVertex3f(columnsX,columnsY,0);

	glNormal3f(0.0f, 0.0f, 1.0f);
	glVertex3f(columnsX,0,0);
	glEnd();
	
}

void HeightField::draw() {

	float shininess[] = {70};
	glMaterialfv(GL_FRONT, GL_SHININESS, shininess);

	Vector3f normal;
	Vector3f temp1;
	Vector3f xTop;
	Vector3f yTop;

	glBindTexture( GL_TEXTURE_2D, textureHandle );
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);

	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
		//glRotatef(95, 0, 0, 1);
		//glScalef(1000000000.1,1000000000.1, 1000000000.1);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);

	glEnable(GL_TEXTURE_2D);

	for (int x = 0; x < columnsX - 3; x++) {
		for (int y = 0; y < columnsY - 3; y++) {

			if (forceField[x][y] > 0) { // Use to be velocity colored ForceField is -180 out of phase so switch signs! 
				//glColor3f(0.3f, 0.2f, 1.0f);  // Valleys are blue and deceleration (mountains) or white
			} else if (forceField[x][y] < 0) {
				//glColor3f(0.9f, 0.9f, 1.0f);
			} else {
				//glColor3f(0.5f, 0.5f, 0.5f);
			}

			glBegin(GL_QUADS);
			
				temp1.x = x; 
				temp1.y = y;
				temp1.z = heightField[x][y];
			
				xTop.x = (x + 1) - temp1.x;
				xTop.y = (y) - temp1.y;
				xTop.z = (heightField[x + 1][y]) - temp1.z;

				yTop.x = x - temp1.x;
				yTop.y = y + 1 - temp1.y;
				yTop.z = (heightField[x][y + 1]) - temp1.z;

				normal = xTop * yTop;
				//normal = normal.inverted();
				normal = normal.normalized();
				normal = normal.inverted();

				glNormal3f(normal.x, normal.y, normal.z);
			glTexCoord2d( (float)x / columnsX, (float)y / columnsY);
			glVertex3f(x, y, heightField[x][y]);

				temp1.x = x; 
				temp1.y = y + 1;
				temp1.z = heightField[x][y + 1];

				xTop.x = x + 1 - temp1.x;
				xTop.y = y + 1 - temp1.y;
				xTop.z = heightField[x + 1][y + 1] - temp1.z;

				yTop.x = x - temp1.x;
				yTop.y = y + 2 - temp1.y;
				yTop.z = heightField[x][y + 2] - temp1.z;

				//otherVertex.x = (x + 1) - temp1.x;
				//otherVertex.y = (y + 1) - temp1.y;
				//otherVertex.z = (heightField[x + 1][y + 1]) - temp1.z;

				//normal = temp1 * otherVertex;
				//normal = normal.inverted();
				//normal = normal.normalized();

				normal = xTop * yTop;
				normal = normal.normalized();
				normal = normal.inverted();

				glNormal3f(normal.x, normal.y, normal.z);
				glTexCoord2d( (float)x / columnsX, (float)(y+1) / columnsY);
			glVertex3f(x, y + 1, heightField[x][y + 1]);


				temp1.x = x + 1; 
				temp1.y = y + 1;
				temp1.z = heightField[x + 1][y + 1];
			
				xTop.x = x + 2 - temp1.x;
				xTop.y = y + 1 - temp1.y;
				xTop.z = heightField[x + 2][y + 1] - temp1.z;

				yTop.x = x + 1 - temp1.x;
				yTop.y = y + 2 - temp1.y;
				yTop.z = heightField[x + 1][y + 2] - temp1.z;

				normal = xTop * yTop;
				normal = normal.normalized();
				normal = normal.inverted();

				glNormal3f(normal.x, normal.y, normal.z);
				glTexCoord2d( (float)(x+1) / columnsX, (float)(y+1) / columnsY);
			glVertex3f(x + 1, y + 1, heightField[x + 1][y + 1]);

				temp1.x = x + 1; 
				temp1.y = y;
				temp1.z = heightField[x + 1][y];

				xTop.x = x + 2 - temp1.x;
				xTop.y = y - temp1.y;
				xTop.z = heightField[x + 2][y] - temp1.z;

				yTop.x = x + 1 - temp1.x;
				yTop.y = y + 1 - temp1.y;
				yTop.z = heightField[x+1][y+1] - temp1.z;

				normal = xTop * yTop;
				normal = normal.normalized();
				normal = normal.inverted();

				glNormal3f(normal.x, normal.y, normal.z);
				glTexCoord2d( (float)(x+1) / columnsX, (float)(y) / columnsY);
			glVertex3f(x + 1, y, heightField[x + 1][y]);
			//glNormal3f(normal.x, normal.y, normal.z);

			glEnd();

			//glColor3f(0.4f, 0.4f, pow( (heightField[x][y] / (initialHeight)) , 5) );
		}
	}

	glDisable(GL_TEXTURE_2D);
	//glDisable(GL_TEXTURE_GEN_S);
	//glDisable(GL_TEXTURE_GEN_T);
}